@@ -162,11 +162,19 @@ and a unique particle id, the last of which encodes the Lagrangian particle coor
 
 These are described in more detail on the :doc:`AbacusSummit Data Model page <summit:data-products>`.
 
+Use the ``unpack_bits`` argument of the ``CompaSOHaloCatalog`` constructor to specify
+which PID bit fields you want unpacked.  Be aware that some of them might use a lot of
+memory; e.g. the Lagrangian positions are three 4-byte floats per subsample particle.
+Also be aware that some of the returned arrays use narrow int dtypes to save memory,
+such as the ``lagr_idx`` field using ``int16``.  It is easy to silently overflow such
+narrow int types; make sure your operations stay within the type width and cast
+if necessary.
+
 Field Subset Loading
 ====================
 Because the ASDF files are column-oriented, it is possible to load just one or a few
 columns (halo catalog fields) rather than the whole file.  This can save huge amounts
-of IO, memory, and CPU time (due to the decompression).  Use the `fields` argument
+of IO, memory, and CPU time (due to the decompression).  Use the ``fields`` argument
 to the ``CompaSOHaloCatalog`` constructor to specify the list of columns you want.
 
 In detail, some columns are stored as ratios to other columns.  For example, ``r90``
@@ -226,7 +234,7 @@ except:
     # Note: this is a temporary solution until blosc is integrated into ASDF, or until we package a pluggable decompressor
     exit('Error: your ASDF installation does not support Blosc compression.  Please install the fork with Blosc support with the following command: "pip install git+https://github.com/lgarrison/asdf.git"')
 
-from .bitpacked import unpack_rvint, unpack_pids, AUXPID
+from . import bitpacked
 
 # Default to 4 decompression threads, or fewer if fewer cores are available
 DEFAULT_BLOSC_THREADS = 4
@@ -258,8 +266,10 @@ class CompaSOHaloCatalog:
         Parameters
         ----------
         path: str or list of str
-            The directory containing the halo files, like ``MySimulation/halos/z1.000``.
-            Or a halo info file, or a list of halo info files.
+            The halo catalog directory, like ``MySimulation/halos/z1.000/``.
+            Or a single halo info file, or a list of halo info files.
+            Will accept ``halo_info`` dirs or "redshift" dirs
+            (e.g. ``z1.000/halo_info/`` or ``z1.000/``).
 
         load_subsamples: bool or str, optional
             Load halo particle subsamples.  True or False may be specified
@@ -282,11 +292,13 @@ class CompaSOHaloCatalog:
 
             #TODO: above velocity units right?
 
-        unpack_bits: bool, optional
+        unpack_bits: bool, or list of str, optional
             Extract information from the PID field of each subsample particle
-            info about its Lagrangian position, whether it is tagged,  and its
+            info about its Lagrangian position, whether it is tagged, and its
             current local density.  If False, only the particle ID part will
-            be extracted.
+            be extracted.  Note that this per-particle information can be large.
+            Can be a list of str, in which case only those fields will be unpacked.
+            Field names are: ('pid', 'lagr_pos', 'tagged', 'density', 'lagr_idx').
             Default: False.
 
         fields: str or list of str, optional
@@ -308,23 +320,38 @@ class CompaSOHaloCatalog:
             # if list, must be all files
             for p in path:
                 if os.path.exists(p) and not isfile(p):
-                    raise ValueError(f'If passing a list of paths, all paths must be files. Path "{p}" in list is not a file.')
+                    raise ValueError(f'If passing a list of paths, all paths must be files, not dirs. Path "{p}" is not a file.')
 
         for p in path:
             if not os.path.exists(p):
                 raise FileNotFoundError(f'Path "{p}" does not exist!')
 
         path = [abspath(p) for p in path]
+        
+        # Allow users to pass halo_info dirs, even though redshift dirs remain canoncial
+        for i,p in enumerate(path):
+            if basename(p) == 'halo_info':
+                path[i] = abspath(pjoin(p,os.pardir))
     
         # Can't mix files from different catalogs!
         if isfile(path[0]):
             self.groupdir = dirname(dirname(path[0]))
             for p in path:
-                assert samefile(self.groupdir, dirname(dirname(p)))
+                if not samefile(self.groupdir, dirname(dirname(p))):
+                    raise ValueError("Can't mix files from different catalogs!")
                 halo_fns = path  # path is list of one or more files
+                
+            for i,p in enumerate(path):
+                for j,q in enumerate(path[i+1:]):
+                    if samefile(p,q):
+                        raise ValueError(f'Cannot pass duplicate halo_info files! Found duplicate "{p}" and at indices {i} and {i+j}')
+                
         else:
             self.groupdir = path[0]  # path is a singlet of one dir
-            halo_fns = sorted(glob(pjoin(self.groupdir, 'halo_info', 'halo_info_*')))
+            globpat = pjoin(self.groupdir, 'halo_info', 'halo_info_*')
+            halo_fns = sorted(glob(globpat))
+            if len(halo_fns) == 0:
+                raise FileNotFoundError(f'No halo_info files found! Search pattern was: "{globpat}"')
         del path  # use groupdir and halo_fns
 
         self.chunk_inds = np.array([int(hfn.split('_')[-1].strip('.asdf')) for hfn in halo_fns])
@@ -356,6 +383,17 @@ class CompaSOHaloCatalog:
                 self.load_pidrv = ['pid','rv']
         del load_subsamples  # use the parsed values
 
+        # validate unpack_bits
+        if type(unpack_bits) is str:
+            unpack_bits = [unpack_bits]
+        if unpack_bits not in (True,False):
+            try:
+                for _f in unpack_bits:
+                    assert _f in bitpacked.PID_FIELDS
+            except:
+                raise ValueError(f'`unpack_bits` must be True, False, or one of: "{bitpacked.PID_FIELDS}"')
+        
+        # End parameter parsing, begin opening files
 
         # Open the first file, just to grab the header
         with asdf.open(halo_fns[0], lazy_load=True, copy_arrays=False) as af:
@@ -687,7 +725,7 @@ class CompaSOHaloCatalog:
         for i,af in enumerate(pid_AB_afs):
             thisnp = np_per_file[i]
             if not unpack_bits:
-                pids_AB[start:start+thisnp] = af[self.data_key]['packedpid'] & AUXPID
+                pids_AB[start:start+thisnp] = af[self.data_key]['packedpid'] & bitpacked.AUXPID
             else:
                 pids_AB[start:start+thisnp] = af[self.data_key]['packedpid']
             start += thisnp
@@ -695,19 +733,24 @@ class CompaSOHaloCatalog:
         # Could be expensive!  Off by default.  Probably faster ways to implement this.
         if check_pids:
             assert len(np.unique(pids_AB)) == len(pids_AB)
-
-        if unpack_bits:
+            
+        if unpack_bits:  # anything to unpack?
+            # TODO: eventually, unpacking could be done on each file to save memory, as we do with the rvint
+            if unpack_bits is True:
+                unpack_which = {_f:True for _f in bitpacked.PID_FIELDS}
+            else:  # truthy but not True, like a list
+                unpack_which = {_f:True for _f in unpack_bits}
+                
             # unpack_pids will do unit conversion if requested
             unpackbox = self.header['BoxSize'] if self.convert_units else 1.
-            justpid, lagr_pos, tagged, density = unpack_pids(pids_AB, unpackbox, self.header['ppd'])
-            self.subsamples.add_column(lagr_pos, name='lagr_pos', copy=False)
-            self.subsamples.add_column(tagged, name='tagged', copy=False)
-            self.subsamples.add_column(density, name='density', copy=False)
-            self.subsamples.add_column(justpid, name='pid', copy=False)
-            #self.subsamples.add_column(pids_AB, name='packedpid', copy=False)
+            
+            unpacked_arrays = bitpacked.unpack_pids(pids_AB, box=unpackbox, ppd=self.header['ppd'], **unpack_which)
+            
+            for name in unpacked_arrays:
+                self.subsamples.add_column(unpacked_arrays[name], name=name, copy=False)
         else:
             self.subsamples.add_column(pids_AB, name='pid', copy=False)
-        
+            
             
     def _load_RVs(self, N_halo_per_file):
         
@@ -721,7 +764,7 @@ class CompaSOHaloCatalog:
             particles_AB[start:start+thisnp] = af[self.data_key]['rvint']
             start += thisnp
         unpackbox = self.header['BoxSize'] if self.convert_units else 1.
-        ppos_AB, pvel_AB = unpack_rvint(particles_AB, unpackbox)
+        ppos_AB, pvel_AB = bitpacked.unpack_rvint(particles_AB, unpackbox)
 
         self.subsamples.add_column(ppos_AB, name='pos', copy=False)
         self.subsamples.add_column(pvel_AB, name='vel', copy=False)
